use crate::{
    prelude::*,
    runtime,
    types,
    util::binary,
};
use crate::interface::*;

pub fn encode_tag(type_id:usize) -> Vec<u8>
{
    let mut tid = type_id;
    let mut result = Vec::<u8>::new();
    let mut remaining :usize = 1;
    while remaining > 0 {
        result.append(&mut binary::pack_natural(unsafe {runtime::type_key(tid)} as u64));

        remaining -= 1;
        if remaining == 0 {
            remaining += unsafe {runtime::type_hasinner(tid)};
        }
        tid = unsafe {runtime::type_inner(tid)};
    }
    return result;
}

pub fn encode_data(addr:Ref) -> Vec<u8>
{
    match unsafe {runtime::type_key(addr.class)} {
        types::TYPE_VAR => {
            encode(Var::__from_ref(addr).get_any().__to_ref())
        }
        types::TYPE_BOOLEAN => {
            binary::pack_natural(unsafe {runtime::bool_get(addr)} as u64)
        }
        types::TYPE_NATURAL => {
            binary::pack_natural(unsafe {runtime::natural_get(addr)})
        }
        types::TYPE_INTEGER => {
            binary::pack_integer(unsafe {runtime::integer_get(addr)})
        }
        /*types::TYPE_DECIMAL => {
            let mantissa = unsafe {type_innerkey(addr.class)};
            let raw = Decimal::try_from(addr).unwrap().get_raw();
            let whole = raw >> mantissa;
            let fract = (raw as u64 & ((1u64 << mantissa) - 1)).reverse_bits() >> (u64::BITS as usize - mantissa);
            result.append(&mut &mut binary::pack_integer(whole));
            result.append(&mut &mut binary::pack_natural(fract));
        }*/
        /*types::TYPE_SIGNIFICANT => {
            result.append(&mut vec![0]);
        }*/
        types::TYPE_BLOCK => {
            let length = unsafe {runtime::type_innerkey(addr.class)};
            let mut result = Vec::new();
            for i in 0..length {
                result.push(unsafe {runtime::block_get(addr, i)});
            }
            result
        }
        types::TYPE_SEQUENCE => { 
            let length = unsafe {runtime::sequence_length(addr)};

            let mut result = binary::pack_natural(length as u64);
            for i in 0..length {
                result.push(unsafe {runtime::sequence_get(addr, i)});
            }
            result
        }
        types::TYPE_ARRAY => {
            let len_type = unsafe {runtime::type_inner(addr.class)};
            let length = unsafe {runtime::type_key(len_type)};

            let mut result = Vec::new();
            for i in 0..length {
                let refr = unsafe {runtime::array_at(addr, i)};

                if unsafe {runtime::type_key(refr.class)} == types::TYPE_VAR {
                    result.append(&mut encode(unsafe {runtime::var_get(refr)}));
                }
                else {
                    result.append(&mut encode_data(refr));
                }
            }
            result
        }
        types::TYPE_LIST => {
            let length = unsafe {runtime::list_length(addr)};

            let mut result = Vec::new();
            for i in 0..length {
                let refr = unsafe {runtime::list_at(addr, i)};

                if unsafe {runtime::type_key(refr.class)} == types::TYPE_VAR {
                    result.append(&mut encode(unsafe {runtime::var_get(refr)}));
                }
                else {
                    result.append(&mut encode_data(refr));
                }
            }
            result
        }
        types::TYPE_SPARSE => {
            let result = Vec::new();
            /*let data = crate::Sparse::try_from(addr).unwrap();

            let header_length = unsafe {runtime::sparse_header_length(addr)};

            result.append(&mut binary::pack_natural(header_length as u64));

            for i in 0..header_length {
                let header = unsafe {runtime::sparse_header_data(addr, i)};
                result.append(&mut binary::pack_natural(header.start as u64));
                result.append(&mut binary::pack_natural(header.length as u64));
                result.append(&mut binary::pack_natural(header.index as u64));
            }

            for i in 0..data.length() {
                if unsafe {type_innerkey(addr.class)} == types::TYPE_VAR {
                    result.append(&mut encode(data.at(i)));
                }
                else {
                    result.append(&mut encode_data(data.at(i)));
                }
            }*/
            result
        }
        types::TYPE_RECORD => {
            let _schema = unsafe {runtime::type_innerkey(addr.class)};
            /*let length = unsafe {runtime::schema_length(schema)};

            let data = Record::try_from(addr).unwrap();
            for i in 0..data.length() {
                if kind(data.kindof(i)) == types::TYPE_VAR {
                    result.append(&mut encode(data.at(i)));
                }
                else {
                    result.append(&mut encode_data(data.at(i)));
                }
            }*/
            Vec::new()
        }
        types::TYPE_SCHEMA => {
            /*let data = Schema::try_from(addr).unwrap();

            result.append(&mut binary::pack_natural(data.length() as u64));
            
            for i in 0..data.length() {
                result.append(&mut encode_tag(data.get(i).unwrap()));

                let mut key = Vec::from(data.keyof(i).unwrap().as_bytes());
                result.append(&mut binary::pack_natural(key.len() as u64));
                result.append(&mut key);
            }*/
            Vec::new()
        }
        _ => { panic!("szun encode not implemented"); }
    }
}

pub fn encode(addr:Ref) -> Vec<u8>
{
    let mut result = encode_tag(addr.class);
    result.append(&mut encode_data(addr));
    return result;
}

pub fn decode_tag(data:&Vec<u8>, index:&mut usize) -> Result<usize,SzunError>
{
    let mut tags = Vec::<usize>::new();
    let mut remaining :usize = 1;
    while remaining > 0 && *index < data.len() {
        let kind = match binary::unpack_natural(&data, index) {
            Ok(id) => id as usize,
            Err(e) => return Err(e),
        };
        tags.push(kind);

        remaining -= 1;
        if remaining == 0 {
            remaining += unsafe {runtime::kind_hasinner(kind)};
        }
    }
    let mut type_id = 0;
    for i in (0..tags.len()).rev() {
        type_id = unsafe {runtime::type_outer(type_id, tags[i])};
    }

    return if remaining == 0 { Ok(type_id) }
    else { Err(SzunError::BadData) }
}

pub fn decode_data(_data:&Vec<u8>, type_id:usize, _index:&mut usize) -> Result<Any,()>
{
    match unsafe {runtime::type_key(type_id)} {
        /*types::TYPE_VAR => {
            return match decode(data, index) {
                Ok(value) => {
                    Ok(Object::Var(Var::with(value.get())))
                }
                Err(_) => Err(()),
            }
        }
        types::TYPE_BOOLEAN => {
            return Ok(Object::Boolean(Boolean::with(unpack_natural(data, index) == 1)));
        }
        types::TYPE_NATURAL => {
            return Ok(Object::Natural(Natural::with(unpack_natural(data, index))));
        }
        types::TYPE_INTEGER => {
            return Ok(Object::Integer(Integer::with(unpack_integer(data, index))));
        }
        types::TYPE_DECIMAL => {
            let mantissa = unsafe {type_innerkey(type_id)};
            let whole = unpack_integer(data, index);
            let fract = unpack_natural(data, index).reverse_bits() >> (u64::BITS as usize - mantissa);
            let raw = (whole << mantissa) | fract as i64;

            return Ok(Object::Decimal(Decimal::with_raw(unsafe {type_innerkey(type_id)}, raw)));
        }
        types::TYPE_SIGNIFICANT => {
            return Err(())
        }
        types::TYPE_BLOCK => {
            let size = unsafe {type_innerkey(type_id)};
            let mut bytes = Vec::<u8>::with_capacity(size);
            if *index + size <= data.len() {
                for _ in 0..size {
                    bytes.push(data[*index]);
                    *index += 1;
                }
            }
            return Ok(Object::Block(Block::with(size, bytes)));
        }
        types::TYPE_SEQUENCE => {
            let size = unpack_natural(data, index) as usize;
            let mut bytes = Vec::<u8>::with_capacity(size);
            if *index + size <= data.len() {
                for _ in 0..size {
                    bytes.push(data[*index]);
                    *index += 1;
                }
            }
            return Ok(Object::Sequence(Sequence::with_raw(bytes)));
        }
        types::TYPE_ARRAY => {
            let length = unsafe {type_innerkey(type_id)};
            let inner = unsafe {type_inner(type_inner(type_id))};

            let mut result = Array::new(length, inner);
            for i in 0..length {
                match if unsafe {type_key(inner)} == types::TYPE_VAR {
                    decode(data, index)
                } else {
                    decode_data(data, inner, index)
                } {
                    Ok(data) => { result.set(i, data.get()); }
                    Err(_) => { return Err(()); }
                }
            }

            return Ok(Object::Array(result));
        }
        types::TYPE_LIST => {
            let inner = unsafe {type_inner(type_id)};
            let length = binary::unpack_natural(data, index) as usize;

            let mut result = List::new(inner);
            result.reserve(length);

            for _ in 0..length {
                match if unsafe {type_key(inner)} == types::TYPE_VAR {
                    decode(data, index)
                } else {
                    decode_data(data, inner, index)
                } {
                    Ok(data) => { result.append(data.get()); }
                    Err(_) => { return Err(()); }
                }
            }

            return Ok(Object::List(result));
        }
        types::TYPE_SPARSE => {
            let inner = unsafe {type_inner(type_id)};
            let mut result = crate::Sparse::new(inner);

            let header_length = binary::unpack_natural(data, index) as usize;
            let mut headers = Vec::<SparseHeader>::with_capacity(header_length);
            
            for _ in 0..header_length {
                headers.push(runtime::SparseHeader {
                    start:binary::unpack_natural(data, index) as usize,
                    length:binary::unpack_natural(data, index) as usize,
                    index:binary::unpack_natural(data, index) as usize,
                });
            }

            for header in headers {
                for i in 0..header.length {
                    match if unsafe {type_key(inner)} == types::TYPE_VAR {
                        decode(data, index)
                    } else {
                        decode_data(data, inner, index)
                    } {
                        Ok(data) => {
                            result.set(header.start + i, data.get());
                        }
                        Err(_) => { return Err(()); }
                    }
                }
            }

            return Ok(Object::Sparse(result));
        }
        types::TYPE_RECORD => {
            return match Record::new(unsafe {type_innerkey(type_id)}) {
                Ok(mut value) => {
                    for i in 0..value.length() {
                        match decode_data(data, value.kindof(i), index) {
                            Ok(refer) => {
                                value.set_at(i, refer.get());
                            }
                            Err(_) => return Err(())
                        }
                    }
                    Ok(Object::Record(value))
                }
                Err(_) => Err(()),
            }
        }
        types::TYPE_SCHEMA => {
            let mut result = Schema::new();

            let length = binary::unpack_natural(data, index) as usize;

            for _ in 0..length {
                let class :usize;
                match decode_tag(data, index) {
                    Ok(tag) => { class = tag; }
                    Err(_) => { return Err(()); }
                }

                let key_length = binary::unpack_natural(data, index) as usize;
                let mut key_bytes = Vec::<u8>::with_capacity(key_length);
                for _ in 0..key_length {
                    if *index < data.len() {
                        key_bytes.push(data[*index]);
                        *index += 1;
                    }
                    else {
                        return Err(());
                    }
                }
                let key = String::from_utf8(key_bytes).unwrap();

                result.assign(&key, class);
            }

            return Ok(Any::__from_ref(result));
        }*/
        _ => { panic!("szun decode not implemented"); }
    }
}

pub fn decode(data:&Vec<u8>, index:&mut usize) -> Result<Any,()>
{
    return match decode_tag(data, index) {
        Ok(type_id) => decode_data(data, type_id, index),
        _ => Err(()),
    }
}
