use crate::{
    prelude::*, types,
    runtime,
};

pub struct Block<const N :usize> {
    address:Ref,
}
impl<const N :usize> Block<N> {
    pub fn new() -> Self
    {
        Self { address:unsafe {runtime::acquire(Self::type_id())} }
    }

    pub fn with(mut self, data:&[u8; N]) -> Self
    {
        self.set(data);
        self
    }

    pub fn set(&mut self, data:&[u8; N])
    {
        for i in 0..data.len() {
            unsafe {runtime::block_set(self.address.clone(), i, data[i])};
        }
    }

    pub fn get(&self) -> [u8; N]
    {
        let mut data = [0; N];
        for i in 0..data.len() {
            data[i] = unsafe {runtime::block_get(self.address.clone(), i)};
        }
        data
    }

    pub fn set_at(&mut self, index:usize, data:u8)
    {
        unsafe {runtime::block_set(self.address.clone(), index, data)}
    }

    pub fn get_at(&self, index:usize) -> u8
    {
        unsafe {runtime::block_get(self.address.clone(), index)}
    }

    pub fn set_unsigned<T: num::Num + num::NumCast>(&mut self, data:T)
    {
        let mut value :u128 = num::NumCast::from(data).unwrap();
        let n = N - 1;
        for i in 0..N {
            unsafe {runtime::block_set(self.address.clone(), n - i, (value & 0xFF) as u8)};
            value >>= 8;
        }
    }

    pub fn get_unsigned<T: num::Num + num::NumCast>(&mut self) -> T
    {
        let mut value :u128 = 0;
        for i in 0..N {
            value <<= 8;
            value |= unsafe {runtime::block_get(self.address.clone(), i)} as u128;
        }
        num::NumCast::from(value).unwrap()
    }

    pub fn set_bits(&mut self, _index:usize, _length:usize, _data:usize)
    {
        panic!("block.set_bits not implemented");
        //unsafe {block_set_bits(self.address.clone(), index, length, data)}
    }

    pub fn get_bits(&self, _index:usize, _length:usize) -> usize
    {
        panic!("block.get_bits not implemented");
        //unsafe {block_get_bits(self.address.clone(), index, length)}
    }

    pub fn len(&self) -> usize
    {
        N
    }
}
impl<const N :usize> SzunType for Block<N> {
    fn type_id() -> usize
    {
        let inner = unsafe {runtime::type_outer(0, N)};
        unsafe {runtime::type_outer(inner, types::TYPE_BLOCK)}
    }

    fn compatible<T: SzunType>(&self, _:&T) -> bool { false }

    fn __to_ref(&self) -> Ref { self.address.clone() }
    fn __from_ref(refr:Ref) -> Self { Self { address:refr } }
}
impl<const N :usize> Clone for Block<N> {
    fn clone(&self) -> Self
    {
        Self::__from_ref(unsafe {runtime::clone(self.address.clone())})
    }
}
impl<const N :usize> TryFrom<Any> for Block<N> {
    type Error = SzunError;
    fn try_from(obj: Any) -> Result<Self, Self::Error>
    {
        if obj.type_id() == Self::type_id() {
            Ok(Self::__from_ref(obj.__to_ref()))
        } else { Err(SzunError::MismatchedType) }
    }
}
impl<const N :usize> Into<Any> for Block<N> { fn into(self) -> Any { Any::__from_ref(self.address.clone()) } }
impl<const N :usize> Drop for Block<N> {
    fn drop(&mut self)
    {
        unsafe {runtime::release(self.address.clone())};
    }
}

pub fn block<const N :usize>(data:&[u8; N]) -> Block<N>
{
    Block::new().with(data)
}
