use std::marker::PhantomData;
use crate::{
    prelude::*,
    runtime,
    types,
};

pub struct List<T: SzunType> {
    address:Ref,
    _data:PhantomData<T>,
}
impl<T: SzunType> List<T> {
    pub fn new() -> Self
    {
        Self {
            address:unsafe {runtime::acquire(Self::type_id())},
            _data:PhantomData
        }
    }

    pub fn insert<D: SzunType>(&mut self, index:usize, data:&D)
    {
        if self.compatible(data) {
            unsafe {runtime::list_insert(self.address.clone(), index, data.__to_ref())};
        } else {
            panic!("szun mismatched type");
        }
    }

    pub fn push<D: SzunType>(&mut self, data:&D)
    {
        if self.compatible(data) {
            unsafe {runtime::list_push(self.address.clone(), data.__to_ref())};
        } else {
            panic!("szun mismatched type");
        }
    }

    pub fn pop(&mut self) -> Option<T>
    {
        let refr = unsafe {runtime::list_back(self.address.clone())};
        if refr.is_some() {
            let result = Some(T::__from_ref(unsafe {runtime::clone(refr)}));
            unsafe {runtime::list_pop(self.address.clone())};
            result
        } else { None }
    }

    pub fn set<D: SzunType>(&mut self, index:usize, data:&D)
    {
        if self.compatible(data) {
            unsafe {runtime::list_set(self.address.clone(), index, data.__to_ref())};
        } else {
            panic!("szun mismatched type");
        }
    }

    pub fn clear(&mut self)
    {
        unsafe {runtime::list_clear(self.address.clone())};
    }

    pub fn reserve(&mut self, length:usize)
    {
        unsafe {runtime::list_reserve(self.address.clone(), length)};
    }

    pub fn get(&mut self, index:usize) -> T
    {
        let refr = unsafe {runtime::list_at(self.address.clone(), index)};
        if refr.is_some() {
            T::__from_ref(refr)
        } else { panic!("index out of bounds: {} / {}", index, self.len()) }
    }

    pub fn try_get(&mut self, index:usize) -> Option<T>
    {
        let refr = unsafe {runtime::list_at(self.address.clone(), index)};
        if refr.is_some() {
            Some(T::__from_ref(refr))
        } else { None }
    }

    pub fn len(&self) -> usize
    {
        unsafe {runtime::list_length(self.address.clone())}
    }

    pub fn capacity(&self) -> usize
    {
        unsafe {runtime::list_capacity(self.address.clone())}
    }
}
impl<T: SzunType> SzunType for List<T> {
    fn type_id() -> usize
    {
        unsafe {runtime::type_outer(T::type_id(), types::TYPE_LIST)}
    }

    fn compatible<R: SzunType>(&self, _:&R) -> bool
    {
        R::type_id() == T::type_id() || unsafe {runtime::type_key(T::type_id())} == types::TYPE_VAR
    }

    fn __to_ref(&self) -> Ref { self.address.clone() }
    fn __from_ref(address:Ref) -> Self { Self { address:address, _data:PhantomData } }
}
impl<T: SzunType> Clone for List<T> {
    fn clone(&self) -> Self { Self::__from_ref(unsafe {runtime::clone(self.address.clone())}) }
}
impl<T: SzunType> TryFrom<Any> for List<T> {
    type Error = SzunError;
    fn try_from(obj: Any) -> Result<Self, Self::Error>
    {
        if obj.type_id() == Self::type_id() {
            Ok(Self::__from_ref(obj.__to_ref()))
        } else { Err(SzunError::MismatchedType) }
    }
}
impl<T: SzunType> Into<Any> for List<T> { fn into(self) -> Any { Any::__from_ref(self.address.clone()) } }
impl<T: SzunType> Drop for List<T> {
    fn drop(&mut self) { unsafe {runtime::release(self.address.clone())}; }
}
