use crate::{
    prelude::*, types,
    runtime,
};

pub struct Sequence {
    pub(crate) address:Ref,
}
impl Sequence {
    pub fn new() -> Self
    {
        Self { address:unsafe {runtime::acquire(Self::type_id())} }
    }

    pub fn with(mut self, data:Vec<u8>) -> Self
    {
        self.set(data);
        self
    }

    pub fn with_str(mut self, data:&str) -> Self
    {
        self.set_str(data);
        self
    }

    pub fn set(&mut self, data:Vec<u8>)
    {
        unsafe {runtime::sequence_resize(self.address.clone(), data.len())};
        for i in 0..data.len() {
            unsafe {runtime::sequence_set(self.address.clone(), i, data[i])};
        }
    }

    pub fn set_str(&mut self, data:&str)
    {
        let data = data.as_bytes();
        unsafe {runtime::sequence_resize(self.address.clone(), data.len())};
        for i in 0..data.len() {
            unsafe {runtime::sequence_set(self.address.clone(), i, data[i])};
        }
    }

    pub fn get(&self) -> Vec<u8>
    {
        let mut data = Vec::new();
        for i in 0..data.len() {
            data.push(unsafe {runtime::sequence_get(self.address.clone(), i)});
        }
        data
    }

    pub fn get_str(&self) -> String
    {
        let mut data = Vec::new();
        for i in 0..self.len() {
            data.push(unsafe {runtime::sequence_get(self.address.clone(), i)});
        }
        String::from_utf8(data).unwrap()
    }

    pub fn set_at(&mut self, index:usize, data:u8)
    {
        unsafe {runtime::sequence_set(self.address.clone(), index, data)}
    }

    pub fn get_at(&self, index:usize) -> u8
    {
        unsafe {runtime::sequence_get(self.address.clone(), index)}
    }

    pub fn len(&self) -> usize
    {
        unsafe {runtime::sequence_length(self.address.clone())}
    }
}
impl SzunType for Sequence {
    fn type_id() -> usize { unsafe {runtime::type_outer(0, types::TYPE_SEQUENCE)} }

    fn compatible<T: SzunType>(&self, _:&T) -> bool { false }

    fn __to_ref(&self) -> Ref { self.address.clone() }
    fn __from_ref(refr:Ref) -> Self { Self { address:refr } }
}
impl Clone for Sequence {
    fn clone(&self) -> Self { Self::__from_ref(unsafe {runtime::clone(self.address.clone())}) }
}
impl TryFrom<Any> for Sequence {
    type Error = SzunError;
    fn try_from(obj: Any) -> Result<Self, Self::Error>
    {
        if obj.type_id() == Self::type_id() {
            Ok(Self::__from_ref(obj.__to_ref()))
        } else { Err(SzunError::MismatchedType) }
    }
}
impl Into<Any> for Sequence { fn into(self) -> Any { Any::__from_ref(self.address.clone()) } }
impl Drop for Sequence {
    fn drop(&mut self)
    {
        unsafe {runtime::release(self.address.clone())};
    }
}

pub fn sequence(data:Vec<u8>) -> Sequence
{
    Sequence::new().with(data)
}

pub fn string(data:&str) -> Sequence
{
    Sequence::new().with_str(data)
}
