use std::marker::PhantomData;
use crate::{
    prelude::*,
    runtime,
    types,
};

pub struct Sparse<T: SzunType> {
    address:Ref,
    _data:PhantomData<T>,
}
impl<T: SzunType> Sparse<T> {
    pub fn new() -> Self
    {
        Self {
            address:unsafe {runtime::acquire(Self::type_id())},
            _data:PhantomData
        }
    }

    pub fn set<D: SzunType>(&mut self, index:usize, data:&D)
    {
        if self.compatible(data) {
            unsafe {runtime::sparse_set(self.address.clone(), index, data.__to_ref())};
        }
    }
    
    pub fn unset<D: SzunType>(&mut self, index:usize, data:&D)
    {
        if self.compatible(data) {
            unsafe {runtime::sparse_unset(self.address.clone(), index)};
        }
    }

    pub fn clear(&mut self)
    {
        unsafe {runtime::sparse_clear(self.address.clone())};
    }

    pub fn get(&mut self, index:usize) -> T
    {
        let refr = unsafe {runtime::sparse_get(self.address.clone(), index)};
        if refr.is_some() {
            T::__from_ref(refr)
        } else { panic!("index out of bounds: {} / {}", index, self.len()) }
    }
    
    pub fn try_get(&mut self, index:usize) -> Option<T>
    {
        let refr = unsafe {runtime::sparse_get(self.address.clone(), index)};
        if refr.is_some() {
            Some(T::__from_ref(refr))
        } else { None }
    }

    pub fn at(&mut self, index:usize) -> T
    {
        let refr = unsafe {runtime::sparse_at(self.address.clone(), index)};
        if refr.is_some() {
            T::__from_ref(refr)
        } else { panic!("index out of bounds: {} / {}", index, self.len()) }
    }

    pub fn first(&self) -> usize
    {
        unsafe {runtime::sparse_first(self.address.clone())}
    }

    pub fn last(&self) -> usize
    {
        unsafe {runtime::sparse_last(self.address.clone())}
    }

    pub fn next(&self) -> usize
    {
        unsafe {runtime::sparse_next(self.address.clone())}
    }

    pub fn len(&self) -> usize
    {
        unsafe {runtime::sparse_length(self.address.clone())}
    }
}
impl<T: SzunType> SzunType for Sparse<T> {
    fn type_id() -> usize { unsafe {runtime::type_outer(T::type_id(), types::TYPE_SPARSE)} }

    fn compatible<R: SzunType>(&self, _:&R) -> bool
    {
        R::type_id() == T::type_id() || unsafe {runtime::type_key(T::type_id())} == types::TYPE_VAR
    }

    fn __to_ref(&self) -> Ref { self.address.clone() }
    fn __from_ref(address:Ref) -> Self { Self { address:address, _data:PhantomData } }
}
impl<T: SzunType> Clone for Sparse<T> {
    fn clone(&self) -> Self { Self::__from_ref(unsafe {runtime::clone(self.address.clone())}) }
}
impl<T: SzunType> TryFrom<Any> for Sparse<T> {
    type Error = SzunError;
    fn try_from(obj: Any) -> Result<Self, Self::Error>
    {
        if obj.type_id() == Self::type_id() {
            Ok(Self::__from_ref(obj.__to_ref()))
        } else { Err(SzunError::MismatchedType) }
    }
}
impl<T: SzunType> Into<Any> for Sparse<T> { fn into(self) -> Any { Any::__from_ref(self.address.clone()) } }
impl<T: SzunType> Drop for Sparse<T> {
    fn drop(&mut self)
    {
        unsafe {runtime::release(self.address.clone())};
    }
}

impl<T: SzunType> TryFrom<Ref> for Sparse<T> {
    type Error = SzunError;

    fn try_from(addr: Ref) -> Result<Self, Self::Error>
    {
        if addr.class == Self::type_id() {
            Ok(Self { address:addr, _data:PhantomData })
        } else { Err(SzunError::MismatchedType) }
    }
}
impl<T: SzunType> Into<Ref> for Sparse<T> {
    fn into(self) -> Ref { self.address.clone() }
}

