use crate::{
    prelude::*,
    runtime,
    types,
    Any,
};

pub struct Var {
    address:Ref,
}
impl Var {
    pub fn new() -> Self
    {
        Self { address:unsafe {runtime::acquire(Self::type_id())} }
    }

    pub fn with<T: SzunType>(mut self, data:&T) -> Self
    {
        self.set(data);
        self
    }

    pub fn set<T: SzunType>(&mut self, data:&T)
    {
        unsafe {runtime::var_set(self.address.clone(), data.__to_ref())}
    }
    
    pub fn get<T: SzunType>(&self) -> T
    {
        let refr = unsafe {runtime::var_get(self.address.clone())};
        if unsafe {runtime::type_sanitize(refr.class)} == T::type_id() {
            T::__from_ref(refr)
        } else { panic!("mismatched szun types"); }
    }

    pub fn get_any(&self) -> Any
    {
        let refr = unsafe {runtime::var_get(self.address.clone())};
        Any::__from_ref(refr)
    }

    pub fn try_get<T: SzunType>(&self) -> Result<T,SzunError>
    {
        let refr = unsafe {runtime::var_get(self.address.clone())};
        if unsafe {runtime::type_sanitize(refr.class)} == T::type_id() {
            Ok(T::__from_ref(refr))
        } else { Err(SzunError::MismatchedType) }
    }

    pub fn has<T: SzunType>(&self) -> bool
    {
        let refr = unsafe {runtime::var_get(self.address.clone())};
        unsafe {runtime::type_sanitize(refr.class) == T::type_id()}
    }

    pub fn clear(&mut self)
    {
        unsafe {runtime::var_set(self.address.clone(), Ref::null())}
    }
}
impl SzunType for Var {
    fn type_id() -> usize { unsafe {runtime::type_outer(0, types::TYPE_VAR)} }
    
    fn compatible<T: SzunType>(&self, _:&T) -> bool { true }

    fn __to_ref(&self) -> Ref { self.address.clone() }
    fn __from_ref(refr:Ref) -> Self { Self { address:refr } }
}
impl Clone for Var {
    fn clone(&self) -> Self
    {
        Self::__from_ref(unsafe {runtime::clone(self.address.clone())})
    }
}
impl TryFrom<Any> for Var {
    type Error = SzunError;
    fn try_from(obj: Any) -> Result<Self, Self::Error>
    {
        if obj.type_id() == Self::type_id() {
            Ok(Self { address:obj.__to_ref() })
        } else { Err(SzunError::MismatchedType) }
    }
}
impl Into<Any> for Var { fn into(self) -> Any { Any::__from_ref(self.address.clone()) } }
impl Drop for Var {
    fn drop(&mut self)
    {
        unsafe {runtime::release(self.address.clone())};
    }
}

pub fn var<T: SzunType>(data:&T) -> Var
{
    Var::new().with(data)
}
