#include "lib.h"

size_t sparse_header_length(Reference addr)
{
    auto& list = *reinterpret_cast<Type::Sparse*>(addr.address);
    return list.header.length;
}

Type::SparseHeader sparse_header_data(Reference addr, size_t index)
{
    auto& list = *reinterpret_cast<Type::Sparse*>(addr.address);
    Type::SparseHeader header {0};

    if(index < list.header.length) {
        header = *reinterpret_cast<Type::SparseHeader*>(rawlist_cell(list.header, sizeof(Type::SparseHeader), index));
    }

    return header;
}

extern "C" size_t sparse_length(Reference addr)
{
    auto& list = *reinterpret_cast<Type::Sparse*>(addr.address);
    return list.data.length;
}

extern "C" void sparse_clear(Reference addr)
{
    auto& list = *reinterpret_cast<Type::Sparse*>(addr.address);
    size_t inner = type_inner(addr.type);

    Reference ref {0};
    ref.type = inner;
    
    for(size_t i = 0; i < list.data.length; ++i) {
        ref.address = rawlist_cell(list.data, type_size(inner), i);
        drop(ref);
    }

    rawlist_clear(list.data);
    rawlist_clear(list.header);
}

extern "C" size_t sparse_first(Reference addr)
{
    auto& list = *reinterpret_cast<Type::Sparse*>(addr.address);

    if(list.header.length > 0) {
        auto cell = reinterpret_cast<Type::SparseHeader*>(rawlist_cell(list.header, sizeof(Type::SparseHeader), 0));
        if(cell != nullptr) {
            return cell->start;
        }
    }
    return 0;
}

extern "C" size_t sparse_next(Reference addr)
{
    auto& list = *reinterpret_cast<Type::Sparse*>(addr.address);

    if(list.header.length > 0) {
        auto cell = reinterpret_cast<Type::SparseHeader*>(rawlist_cell(list.header, sizeof(Type::SparseHeader), 0));
        if(cell != nullptr) {
            return cell->start + cell->length;
        }
    }
    return 0;
}

extern "C" size_t sparse_last(Reference addr)
{
    auto& list = *reinterpret_cast<Type::Sparse*>(addr.address);

    if(list.header.length > 0) {
        auto cell = reinterpret_cast<Type::SparseHeader*>(rawlist_cell(list.header, sizeof(Type::SparseHeader), list.header.length - 1));
        if(cell != nullptr) {
            return cell->start + cell->length - 1;
        }
    }
    return 0;
}

extern "C" Reference sparse_at(Reference addr, size_t index)
// Returns 
//
{
    Reference result {0};
    auto& list = *reinterpret_cast<Type::Sparse*>(addr.address);
    size_t inner = type_inner(addr.type);

    if(index < list.data.length) {
        auto cell = rawlist_cell(list.data, type_size(inner), index);
        if(cell != nullptr) {
            result.type = inner;
            result.address = cell;
        }

        if(type_key(inner) == Type::TYPE_VARYING) {
            result = var_get(result);
        }
    }

    return refr_make_temporary(result);
}

extern "C" Reference sparse_get(Reference addr, size_t index)
{
    Reference result {0};
    auto& list = *reinterpret_cast<Type::Sparse*>(addr.address);
    size_t inner = type_inner(addr.type);

    size_t bound_lower = 0;
    size_t bound_upper = list.header.length;
    size_t header_index = 0;
    Type::SparseHeader* header = nullptr;

    while(bound_lower != bound_upper) {
        header_index = bound_lower + ((bound_upper - bound_lower) / 2);
        header = reinterpret_cast<Type::SparseHeader*>(rawlist_cell(list.header, sizeof(Type::SparseHeader), header_index));
        
        if(index >= header->start && index < (header->start + header->length)) {
            break;
        }
        else if(index < header->start) {
            bound_upper = header_index;
        }
        else {
            bound_lower = header_index + 1;
        }

        header = nullptr;
    }

    if(header != nullptr) {
        result.type = inner;
        size_t data_index = header->index + (index - header->start);
        result.address = rawlist_cell(list.data, type_size(inner), data_index);

        if(type_key(result.type) == Type::TYPE_VARYING) {
            result = var_get(result);
        }
    }

    return refr_make_temporary(result);
}

extern "C" void sparse_set(Reference addr, size_t index, Reference source)
{
    auto& list = *reinterpret_cast<Type::Sparse*>(addr.address);
    size_t inner = type_inner(addr.type);

    size_t header_index = 0;
    Type::SparseHeader* header = nullptr;
    for(; header_index < list.header.length; ++header_index) {
        header = reinterpret_cast<Type::SparseHeader*>(rawlist_cell(list.header, sizeof(Type::SparseHeader), header_index));
        if(index <= (header->start + header->length)) {
            break;
        }
    }

    Type::SparseHeader header_new;

    Reference new_data {0};
    new_data.type = inner;

    if(header_index == list.header.length) {
        // append header to end
        header_new.start = index;
        header_new.length = 1;
        header_new.index = list.data.length;

        new_data.address = rawlist_insert(list.data, type_size(inner), list.data.length);
        copy(new_data, source);

        auto& h_cell = *reinterpret_cast<Type::SparseHeader*>(rawlist_insert(list.header, sizeof(Type::SparseHeader), list.header.length));
        h_cell = header_new;
    } else {
        if(index < header->start) {
            // insert header before current
            header_new.start = index;
            header_new.length = 1;
            header_new.index = header->index;

            new_data.address = rawlist_insert(list.data, type_size(inner), header->index);
            copy(new_data, source);

            header = reinterpret_cast<Type::SparseHeader*>(rawlist_insert(list.header, sizeof(Type::SparseHeader), header_index));
            *header = header_new;

            for(size_t i = header_index + 1; i < list.header.length; ++i) {
                auto hcell = reinterpret_cast<Type::SparseHeader*>(rawlist_cell(list.header, sizeof(Type::SparseHeader), i));
                hcell->index++;
            }
        } else {
            size_t offset = index - header->start;
            if(offset < header->length) {
                // update existing value
                new_data.address = rawlist_cell(list.data, type_size(inner), header->index + offset);
                drop(new_data);
                copy(new_data, source);
            } else {
                // append value to current 
                new_data.address = rawlist_insert(list.data, type_size(inner), header->index + offset);
                copy(new_data, source);
                header->length++;

                for(size_t i = header_index + 1; i < list.header.length; ++i) {
                    auto hcell = reinterpret_cast<Type::SparseHeader*>(rawlist_cell(list.header, sizeof(Type::SparseHeader), i));
                    hcell->index++;
                }
            }
        }

        // join headers if ranges intersect
        if(header_index + 1 < list.header.length) {
            auto next_header = reinterpret_cast<Type::SparseHeader*>(rawlist_cell(list.header, sizeof(Type::SparseHeader), header_index + 1));
            if(next_header->start == header->start + header->length) {
                header->length += next_header->length;
                rawlist_remove(list.header, sizeof(Type::SparseHeader), header_index + 1);
            }
        }
    }
}

extern "C" void sparse_unset(Reference addr, size_t index)
{
    auto& list = *reinterpret_cast<Type::Sparse*>(addr.address);
    size_t inner = type_inner(addr.type);

    size_t header_index = 0;
    Type::SparseHeader* header = nullptr;
    for(; header_index < list.header.length; ++header_index) {
        header = reinterpret_cast<Type::SparseHeader*>(rawlist_cell(list.header, sizeof(Type::SparseHeader), header_index));
        if(index <= (header->start + header->length)) {
            break;
        }
    }

    Type::SparseHeader header_new;

    Reference cell {0};
    cell.type = inner;

    if(header_index < list.header.length && index >= header->start) {
        size_t offset = index - header->start;
        size_t data_index = header->index + offset;
        if(offset == 0) {
            // shift start of range
            header->start++;
        } else if(offset < header->length - 1) {
            // split header at index
            header_new.start = index + 1;
            header_new.length = header->length - offset - 1;
            header_new.index = header->index + offset + 1;
            header->length = offset + 1;

            auto header_insert = reinterpret_cast<Type::SparseHeader*>(rawlist_insert(list.header, sizeof(Type::SparseHeader), header_index + 1));
            *header_insert = header_new;
        }

        cell.address = rawlist_cell(list.data, type_size(inner), data_index);
        drop(cell);
        rawlist_remove(list.data, type_size(inner), data_index);
        header->length--;

        for(size_t i = header_index + 1; i < list.header.length; ++i) {
            auto hcell = reinterpret_cast<Type::SparseHeader*>(rawlist_cell(list.header, sizeof(Type::SparseHeader), i));
            hcell->index--;
        }

        if(header->length == 0) {
            rawlist_remove(list.header, sizeof(Type::SparseHeader), header_index);
        }
    }
}

/*extern "C" size_t sparse_indexof(Reference addr, size_t index)
{
    auto& list = *reinterpret_cast<Type::Sparse*>(addr.address);

    for(size_t i = 0; i < list.header.length; ++i) {
        auto hcell = reinterpret_cast<Type::SparseHeader*>(rawlist_cell(list.header, sizeof(Type::SparseHeader), i));

        if(index < hcell->length) {
            return hcell->start + index;
        } else {
            index -= hcell->length;
        }
    }

    return 0;
}*/
