uint8_t* allocate(size_t type_id, size_t count)
// Allocates a new block of memory to store n-many instances of type.
//
{
    uint8_t* mem = nullptr;
    size_t size = type_size(type_id) * count;
    if(size > 0) {
        mem = reinterpret_cast<uint8_t*>(malloc(size));
        if(mem != nullptr) {
            memset(mem, 0, size);
        }
        return mem;
    }
    return nullptr;
}

extern "C" Reference acquire(size_t type_id)
// Allocates memory for a type and returns a reference.
//
{
    Reference addr {0};
    addr.address = allocate(type_id, 1);
    if(addr.address != nullptr) {
        addr.type = type_id;
    }
    return addr;
 }

void drop(Reference addr)
// Recursively deletes and zeros an instantiated type.
//
{
    if(addr.address != nullptr) {
        switch(type_key(addr.type)) {
            case Type::TYPE_VARYING: {
                var_set(addr, Reference());
            } break;

            case Type::TYPE_SEQUENCE: {
                sequence_clear(addr);
            } break;

            case Type::TYPE_LIST: {
                list_clear(addr);
            } break;

            case Type::TYPE_SPARSE: {
                sparse_clear(addr);
            } break;

            case Type::TYPE_RECORD: {
                size_t schema_id = type_innerkey(addr.type);
                auto binding = DB_SCHEMA.get(schema_id);
                if(binding != nullptr) {
                    for(size_t i = 0; i < binding->data.size(); ++i) {
                        auto cell = record_cell(addr, i);
                        drop(cell);
                    }
                }
            } break;

            case Type::TYPE_SCHEMA: {
                auto& object = (*reinterpret_cast<Type::Schema*>(addr.address));
                rawlist_clear(object.data);
                rawlist_clear(object.map);
            } break;
        }
        memset(addr.address, 0, type_size(addr.type));
    }
}

extern "C" void release(Reference addr)
// Recursively deletes and frees allocated memory.
//
{
    if(!refr_is_temporary(addr) && addr.address != nullptr) {
        drop(addr);
        free(addr.address);
    }
}

extern "C" Reference clone(Reference src)
// Create a new object with contents equivalent to a source.
//
{
    Reference dst = acquire(src.type);
    copy(dst, src);
    return dst;
}

extern "C" bool copy(Reference dst, Reference src)
// Deep-copy data from a source to a destination.
//
{
    if(src.address != dst.address) {
        Reference source = src;
        Reference destination = dst;
        
        // dereference varying data
        if(type_key(src.type) == Type::TYPE_VARYING) {
            source = *reinterpret_cast<Reference*>(src.address);
        }

        // prepare destination for varying data
        if(type_key(dst.type) == Type::TYPE_VARYING) {
            auto& dest_ref = *reinterpret_cast<Reference*>(dst.address);
            
            // determine if memory can be reused, otherwise free and reallocate
            if(source.type != dest_ref.type) {
                if(dest_ref.address != nullptr) {
                    free(dest_ref.address);
                    dest_ref.type = Type::TYPE_NULL;
                    dest_ref.address = nullptr;
                }
                dest_ref = acquire(source.type);
            }
        }

        // copy data into destination
        if(source.type == destination.type) {
            drop(destination);

            switch(type_key(destination.type)) {
                case Type::TYPE_NULL: { } break;

                case Type::TYPE_SEQUENCE: {
                    auto& src_seq = *reinterpret_cast<Type::Sequence*>(source.address);
                    auto& dst_seq = *reinterpret_cast<Type::Sequence*>(destination.address);

                    rawlist_clear(dst_seq.data);
                    rawlist_reserve(dst_seq.data, sizeof(uint8_t), src_seq.data.length);
                    memcpy(dst_seq.data.data, src_seq.data.data, sizeof(uint8_t) * src_seq.data.length);
                    dst_seq.data.length = src_seq.data.length;
                } break;

                case Type::TYPE_ARRAY: {
                    for(size_t i = 0; i < array_length(source); ++i) {
                        copy(array_cell(destination, i), array_cell(source, i));
                    }
                } break;
                
                case Type::TYPE_LIST: {
                    auto& src_list = *reinterpret_cast<Type::List*>(source.address);
                    auto& dst_list = *reinterpret_cast<Type::List*>(destination.address);

                    rawlist_reserve(dst_list.data, type_size(type_inner(source.type)), src_list.data.capacity);
                    dst_list.data.length = src_list.data.length;

                    for(size_t i = 0; i < src_list.data.length; ++i) {
                        copy(list_at(destination, i), list_at(source, i));
                    }
                } break;

                case Type::TYPE_SPARSE: {
                    auto& src_list = *reinterpret_cast<Type::Sparse*>(source.address);
                    auto& dst_list = *reinterpret_cast<Type::Sparse*>(destination.address);

                    rawlist_reserve(dst_list.header, sizeof(Type::SparseHeader), src_list.header.capacity);
                    dst_list.header.length = src_list.header.length;

                    for(size_t i = 0; i < src_list.header.length; ++i) {
                        Type::SparseHeader& src_header = *reinterpret_cast<Type::SparseHeader*>(rawlist_cell(src_list.header, sizeof(Type::SparseHeader), i));
                        Type::SparseHeader& dst_header = *reinterpret_cast<Type::SparseHeader*>(rawlist_cell(dst_list.header, sizeof(Type::SparseHeader), i));
                        dst_header = src_header;
                    }

                    rawlist_reserve(dst_list.data, type_size(type_inner(source.type)), src_list.data.capacity);
                    dst_list.data.length = src_list.data.length;

                    for(size_t i = 0; i < src_list.data.length; ++i) {
                        copy(list_at(destination, i), list_at(source, i));
                    }
                } break;

                case Type::TYPE_RECORD: {
                    size_t schema_id = type_innerkey(source.type);
                    auto binding = DB_SCHEMA.get(schema_id);
                    if(binding != nullptr) {
                        for(size_t i = 0; i < binding->data.size(); ++i) {
                            auto src_cell = record_cell(source, i);
                            auto dst_cell = record_cell(destination, i);
                            copy(dst_cell, src_cell);
                        }
                    }
                } break;

                case Type::TYPE_SCHEMA: {
                    auto& src_schema = *reinterpret_cast<Type::Schema*>(source.address);
                    auto& dst_schema = *reinterpret_cast<Type::Schema*>(destination.address);

                    rawlist_clear(dst_schema.data);
                    rawlist_clear(dst_schema.map);
                    rawlist_reserve(dst_schema.data, sizeof(size_t), src_schema.data.length);
                    rawlist_reserve(dst_schema.map, sizeof(size_t), src_schema.map.length);

                    for(size_t i = 0; i < src_schema.data.length; ++i) {
                        auto src_cell = reinterpret_cast<size_t*>(rawlist_cell(src_schema.data, sizeof(size_t), i));
                        auto dst_cell = reinterpret_cast<size_t*>(rawlist_cell(dst_schema.data, sizeof(size_t), i));
                        if(src_cell != nullptr && dst_cell != nullptr) {
                            *dst_cell = *src_cell;
                        }
                    }

                    for(size_t i = 0; i < src_schema.map.length; ++i) {
                        auto src_cell = reinterpret_cast<size_t*>(rawlist_cell(src_schema.map, sizeof(size_t), i));
                        auto dst_cell = reinterpret_cast<size_t*>(rawlist_cell(dst_schema.map, sizeof(size_t), i));
                        if(src_cell != nullptr && dst_cell != nullptr) {
                            *dst_cell = *src_cell;
                        }
                    }
                } break;

                default: {
                    memcpy(destination.address, source.address, type_size(source.type));
                } break;
            }

            return true;
        }
    }
    return false;
}

extern "C" bool transfer(Reference dst, Reference src)
// Shallow copy data from a source to a destination and delete the original.
//
{
    if(src.address != dst.address && src.type != 0 && dst.type != 0) {
        Reference source = src;
        Reference destination = dst;
        
        // dereference varying data
        if(type_key(src.type) == Type::TYPE_VARYING) {
            source = *reinterpret_cast<Reference*>(src.address);
        }

        // prepare destination for varying data
        if(type_key(dst.type) == Type::TYPE_VARYING) {
            auto& dest_ref = *reinterpret_cast<Reference*>(dst.address);
            
            // determine if memory can be reused, otherwise free and reallocate
            if(source.type != dest_ref.type) {
                if(dest_ref.address != nullptr) {
                    free(dest_ref.address);
                    dest_ref.type = Type::TYPE_NULL;
                    dest_ref.address = nullptr;
                }
                dest_ref = acquire(source.type);
            }
        }

        // copy data into destination
        if(source.type == destination.type) {
            drop(destination);
            memcpy(destination.address, source.address, type_size(source.type));
            memset(source.address, 0, type_size(source.type));
            return true;
        }
    }
    return false;
}
